package main

import (
	"crypto/hmac"
	"crypto/sha256"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"github.com/gorilla/websocket"
	"io"
	"log"
	"net/http"
	"net/url"
	"os"
	"strings"
	"time"
)

const (
	hostUrl   = "wss://ise-api-sg.xf-yun.com/v2/ise"
	appid     = ""
	apiSecret = ""
	apiKey    = ""
	sub       = "ise"
	ent       = "en_vip"
	category  = "read_sentence"
	text      = "Hello, how are you? i am fine, thank you!"
	file      = "En_tts.pcm"

	StatusFirstFrame    = 0
	StatusContinueFrame = 1
	StatusLastFrame     = 2
)

type IseNewResponseData struct {
	Code    int    `json:"code"`
	Message string `json:"message"`
	Sid     string `json:"sid"`
	Data    Data   `json:"data"`
}

type Data struct {
	Status int    `json:"status"`
	Data   string `json:"data"`
}

func main() {
	fmt.Println("text：" + text)
	authUrl, err := getAuthUrl(hostUrl, apiKey, apiSecret)
	if err != nil {
		log.Fatal(err)
	}
	fmt.Println(authUrl)
	url := strings.Replace(authUrl, "http://", "ws://", 1)
	url = strings.Replace(url, "https://", "wss://", 1)
	fmt.Println("url===>" + url)

	// WebSocket
	c, _, err := websocket.DefaultDialer.Dial(url, nil)
	if err != nil {
		log.Fatal("dial:", err)
	}
	defer c.Close()

	go func() {
		frameSize := 1280
		intervel := 40 * time.Millisecond
		status := StatusFirstFrame

		beginTime := time.Now()

		file, err := os.Open(file)
		if err != nil {
			log.Fatal(err)
		}
		defer file.Close()

		ssb(c)

		buffer := make([]byte, frameSize)

		for {
			len, err := file.Read(buffer)
			if err != nil {
				if err == io.EOF {
					status = StatusLastFrame
				} else {
					log.Fatal(err)
				}
			}

			data := base64.StdEncoding.EncodeToString(buffer[:len])
			switch status {
			case StatusFirstFrame:
				send(c, 1, StatusFirstFrame, data)
				status = StatusContinueFrame
			case StatusContinueFrame:
				send(c, 2, StatusContinueFrame, data)
			case StatusLastFrame:
				send(c, 4, StatusLastFrame, "")
				endTime := time.Now()
				fmt.Printf("time ：%v ms\n", endTime.Sub(beginTime).Milliseconds())
				return
			}

			time.Sleep(intervel)
		}
	}()

	for {
		_, message, err := c.ReadMessage()
		if err != nil {
			log.Println("read:", err)
			return
		}

		var resp IseNewResponseData
		err = json.Unmarshal(message, &resp)
		if err != nil {
			log.Println("json.Unmarshal:", err)
			continue
		}

		if resp.Code != 0 {
			fmt.Printf("code=>%d error=>%s sid=%s\n", resp.Code, resp.Message, resp.Sid)
			return
		}

		if resp.Data.Status == 2 {
			data, err := base64.StdEncoding.DecodeString(resp.Data.Data)
			if err != nil {
				log.Println("base64 decode:", err)
			}
			fmt.Printf("sid:%s result:%s\n", resp.Sid, string(data))
			break
		}
	}
}

func getAuthUrl(hostUrl, apiKey, apiSecret string) (string, error) {
	u, err := url.Parse(hostUrl)
	if err != nil {
		return "", err
	}

	date := time.Now().UTC().Format(http.TimeFormat)
	builder := fmt.Sprintf("host: %s\ndate: %s\nGET %s HTTP/1.1", u.Host, date, u.Path)

	mac := hmac.New(sha256.New, []byte(apiSecret))
	mac.Write([]byte(builder))
	sha := base64.StdEncoding.EncodeToString(mac.Sum(nil))

	authorization := fmt.Sprintf(`api_key="%s", algorithm="hmac-sha256", headers="host date request-line", signature="%s"`, apiKey, sha)
	auth := base64.StdEncoding.EncodeToString([]byte(authorization))

	query := u.Query()
	query.Set("authorization", auth)
	query.Set("date", date)
	query.Set("host", u.Host)

	u.RawQuery = query.Encode()

	return u.String(), nil
}

func ssb(c *websocket.Conn) {
	message := map[string]interface{}{
		"common": map[string]string{
			"app_id": appid,
		},
		"business": map[string]string{
			"category": category,
			"rstcd":    "utf8",
			"sub":      sub,
			"ent":      ent,
			"tte":      "utf-8",
			"cmd":      "ssb",
			"auf":      "audio/L16;rate=16000",
			"aue":      "raw",
			"text":     "\uFEFF" + text,
		},
		"data": map[string]interface{}{
			"status": 0,
			"data":   "",
		},
	}

	jsonData, err := json.Marshal(message)
	if err != nil {
		log.Fatal(err)
	}
	fmt.Println("send:", string(jsonData))
	c.WriteMessage(websocket.TextMessage, jsonData)
}

func send(c *websocket.Conn, aus, status int, data string) {
	message := map[string]interface{}{
		"business": map[string]interface{}{
			"cmd": "auw",
			"aus": aus,
			"aue": "raw",
		},
		"data": map[string]interface{}{
			"status":    status,
			"data":      data,
			"data_type": 1,
			"encoding":  "raw",
		},
	}

	jsonData, err := json.Marshal(message)
	if err != nil {
		log.Fatal(err)
	}
	fmt.Println("send:", string(jsonData))
	c.WriteMessage(websocket.TextMessage, jsonData)
}
